import argparse


def main(x1: float, x2: float) -> float:
    y = (x1**2) + (x2**2)
    return y


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--x1", type=float)
    parser.add_argument("--x2", type=float)

    args, unk_args = parser.parse_known_args()

    print(main(args.x1, args.x2))



import argparse

from aiaccel.config.config import load_config, overwrite_omegaconf_dumper, resolve_inherit
from aiaccel.config.git import collect_git_status_from_config, print_git_status


def main() -> None:
    overwrite_omegaconf_dumper()

    parser = argparse.ArgumentParser()
    parser.add_argument("config", help="Configuration file path")

    args, _ = parser.parse_known_args()
    config = load_config(args.config)
    config = resolve_inherit(config)

    if len(git_status := collect_git_status_from_config(config)) > 0:
        print_git_status(git_status)

        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    main()



from aiaccel.hpo.algorithms.nelder_mead_algorithm import NelderMeadAlgorism, NelderMeadCoefficient, NelderMeadEmptyError

__all__ = [
    "NelderMeadCoefficient",
    "NelderMeadEmptyError",
    "NelderMeadAlgorism",
]



from pathlib import Path

from pytest_mock import MockerFixture

from aiaccel.config.apps import check_git
from aiaccel.config.git import PackageGitStatus


def test_check_git(mocker: MockerFixture) -> None:
    mock_args = mocker.Mock()
    mock_args.config = str(Path(__file__).parent / "test_check_git.yaml")

    mock_argparse = mocker.patch("argparse.ArgumentParser.parse_known_args")
    mock_argparse.return_value = (mock_args, [])

    # Success
    mock_func = mocker.patch("aiaccel.config.apps.check_git.collect_git_status_from_config")
    mock_func.return_value = []

    try:
        check_git.main()
    except SystemExit as e:
        if e.code != 0:
            raise AssertionError() from e

    # Failed
    mock_func = mocker.patch("aiaccel.config.apps.check_git.collect_git_status_from_config")
    mock_func.return_value = [PackageGitStatus("test_package", "test_id", [])]

    try:
        check_git.main()
    except SystemExit as e:
        if e.code != 1:
            raise AssertionError() from e
    else:
        raise AssertionError()



#! /usr/bin/env python3

from argparse import ArgumentParser
import importlib
from pathlib import Path
import pkgutil
import sys


def main() -> None:
    target_module = Path(sys.argv[0]).stem.split("-")[-1]

    package = importlib.import_module(f"aiaccel.{target_module}.apps")

    modules = [name.replace("_", "-") for _, name, ispkg in pkgutil.iter_modules(package.__path__) if not ispkg]
    if not modules:
        raise RuntimeError(f"No apps found in aiaccel.{target_module}.apps")

    parser = ArgumentParser(description=f"Run aiaccel-{target_module} apps.", add_help=False)
    parser.add_argument("command", choices=modules, help="The command to run.")
    args, unk_args = parser.parse_known_args()

    module = importlib.import_module(f"aiaccel.{target_module}.apps.{args.command.replace('-', '_')}")

    sys.argv = [str(module.__file__)] + unk_args
    module.main()


if __name__ == "__main__":
    main()



import numpy as np

import optuna

from aiaccel.hpo.optuna.samplers.nelder_mead_sampler import NelderMeadSampler

search_space = {
    "x": (-10.0, 10.0),
    "y": (-10.0, 10.0),
}


def sphere(trial: optuna.trial.Trial) -> float:
    params = []
    for name, distribution in search_space.items():
        params.append(trial.suggest_float(name, *distribution))

    return float(np.sum(np.asarray(params) ** 2))


if __name__ == "__main__":
    study = optuna.create_study(sampler=NelderMeadSampler(search_space=search_space, seed=42))
    study.enqueue_trial({"x": 1.0, "y": 1.0})
    study.enqueue_trial({"x": 1.0, "y": 2.0})
    study.enqueue_trial({"x": 2.0, "y": 1.0})
    study.optimize(func=sphere, n_trials=100)



from __future__ import annotations

from typing import Any

from collections.abc import Callable, Iterator
from dataclasses import dataclass
from fnmatch import fnmatch

from torch import nn, optim

import lightning as lt
from lightning.pytorch.utilities.types import OptimizerLRSchedulerConfig


@dataclass
class OptimizerConfig:
    """
    Configuration for the optimizer and scheduler in a LightningModule.

    Args:
        optimizer_generator (Callable[..., optim.Optimizer]): A callable that generates the optimizer.
        params_transformer (Callable[..., Iterator[tuple[str, Any]]] | None): A callable that transforms the parameters
            into a format suitable for the optimizer. If None, the parameters are used as is. Defaults to None.
        scheduler_generator (Callable[..., optim.lr_scheduler.LRScheduler] | None):
            A callable that generates the learning rate scheduler. If None, no scheduler is used. Defaults to None.
        scheduler_interval (str | None): The interval at which the scheduler is called. Defaults to "step".
        scheduler_monitor (str | None): The metric to monitor for the scheduler. Defaults to "validation/loss".
    """

    optimizer_generator: Callable[..., optim.Optimizer]
    params_transformer: Callable[..., Iterator[tuple[str, Any]]] | None = None

    scheduler_generator: Callable[..., optim.lr_scheduler.LRScheduler] | None = None
    scheduler_interval: str | None = "step"
    scheduler_monitor: str | None = "validation/loss"


def build_param_groups(
    named_params: Iterator[tuple[str, nn.Parameter]],
    groups: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """
    Build parameter groups for the optimizer based on the provided patterns.

    Args:
        named_params (Iterator[tuple[str, nn.Parameter]]): An iterator of named parameters.
        groups (list[dict[str, Any]]): A list of dictionaries where each dictionary contains
            a "pattern" key that specifies the parameter names to match (``fnmatch``), and other optional keys.

    Example:
    In your config file, you might have:

    .. code-block:: yaml

        optimizer_config:
          _target_: aiaccel.torch.lightning.OptimizerConfig
          optimizer_generator:
            _partial_: True
            _target_: torch.optim.AdamW
            weight_decay: 0.01
          params_transformer:
              _partial_: True
              _target_: aiaccel.torch.lightning.build_param_groups
              groups:
                - pattern: "*bias"
                  lr: 0.01
                - pattern: "*weight"
                  lr: 0.001

    This will create two parameter groups: one for biases with a learning rate of 0.01 and another for weights with
    a learning rate of 0.001.
    """
    remaining = dict(named_params)

    param_groups = []
    for spec in groups:
        matched_params = []
        for target in [spec["pattern"]] if isinstance(spec["pattern"], str) else spec["pattern"]:
            matched_params += [remaining.pop(name) for name in list(remaining.keys()) if fnmatch(name, target)]

        assert len(matched_params) > 0

        param_groups.append({"params": matched_params} | {k: v for k, v in spec.items() if k != "pattern"})

    param_groups.append({"params": list(remaining.values())})

    return param_groups


class OptimizerLightningModule(lt.LightningModule):
    """
    LightningModule subclass for models that use custom optimizers and schedulers.

    Args:
        optimizer_config (OptimizerConfig): Configuration object for the optimizer.

    Attributes:
        optcfg (OptimizerConfig): Configuration object for the optimizer.

    Methods:
        configure_optimizers: Configures the optimizer and scheduler for training.
    """

    def __init__(self, optimizer_config: OptimizerConfig):
        super().__init__()

        self._optimizer_config = optimizer_config

    def configure_optimizers(self) -> optim.Optimizer | OptimizerLRSchedulerConfig:
        """
        Configures the optimizer and scheduler for training.

        Returns:
            Union[optim.Optimizer, OptimizerLRSchedulerConfig]: The optimizer and scheduler configuration.
        """

        params: Iterator[tuple[str, Any]] | Iterator[nn.Parameter]
        if self._optimizer_config.params_transformer is None:
            params = self.parameters()  # just because backward compatibility
        else:
            params = self._optimizer_config.params_transformer(self.named_parameters())

        optimizer = self._optimizer_config.optimizer_generator(params=params)  #

        if self._optimizer_config.scheduler_generator is None:
            return optimizer
        else:
            assert self._optimizer_config.scheduler_interval is not None
            assert self._optimizer_config.scheduler_monitor is not None
            return {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": self._optimizer_config.scheduler_generator(optimizer=optimizer),
                    "interval": self._optimizer_config.scheduler_interval,
                    "monitor": self._optimizer_config.scheduler_monitor,
                },
            }



import json

import lightning


class SaveMetricCallback(lightning.Callback):
    """
    Lightning Callback for save metric in fit ends.

    Args:
        metric_name (str): Metric name to save
        output_path (str): File name to save
    """

    def __init__(self, metric_name: str, output_path: str) -> None:
        super().__init__()
        self.metric_name = metric_name
        self.output_path = output_path

    def on_fit_end(self, trainer: lightning.Trainer, pl_module: lightning.LightningModule) -> None:
        metric_value = trainer.callback_metrics[self.metric_name].item()
        with open(self.output_path, "w") as f:
            json.dump(metric_value, f)



import argparse

from omegaconf import OmegaConf as oc  # noqa: N813

from aiaccel.config.config import load_config, overwrite_omegaconf_dumper, resolve_inherit


def main() -> None:
    overwrite_omegaconf_dumper()

    parser = argparse.ArgumentParser()
    parser.add_argument("config", help="Configuration file path")
    parser.add_argument("key", help="Target key in the configration file")

    args, _ = parser.parse_known_args()
    config = load_config(args.config)
    config = resolve_inherit(config)

    print(oc.select(config, args.key))


if __name__ == "__main__":
    main()



from pathlib import Path

import numpy as np

import torch

import h5py as h5

from aiaccel.torch.datasets.hdf5_dataset import HDF5Dataset, RawHDF5Dataset

# with h5.File(Path(__file__).parent / "test_hdf5_dataset_assets" / "dataset.hdf5", "w") as f:
#     for ii in range(10):
#         g = f.create_group(f"grp{ii}")  # noqa: ERA001
#         g.create_dataset("foo", [2, 3, 4])  # noqa: ERA001
#         g.create_dataset("bar", [5, 6])  # noqa: ERA001

#         g["foo"][:] = np.random.randn(2, 3, 4)  # noqa: ERA001
#         g["bar"][:] = np.random.randn(5, 6)  # noqa: ERA001


def test_raw_hdf5_dataset() -> None:
    hdf5_filename = Path(__file__).parent / "test_hdf5_dataset_assets" / "dataset.hdf5"
    f_hdf5 = h5.File(hdf5_filename)

    dataset = RawHDF5Dataset(hdf5_filename)

    assert len(dataset) == 10
    assert list(dataset.grp_list) == [f"grp{idx}" for idx in range(10)]

    sample = dataset[5]
    assert sorted(sample.keys()) == ["bar", "foo"]
    assert np.array_equal(sample["bar"], f_hdf5["grp5"]["bar"][:])
    assert np.array_equal(sample["foo"], f_hdf5["grp5"]["foo"][:])


def test_hdf5_dataset() -> None:
    hdf5_filename = Path(__file__).parent / "test_hdf5_dataset_assets" / "dataset.hdf5"
    f_hdf5 = h5.File(hdf5_filename)

    dataset = HDF5Dataset(hdf5_filename)

    assert len(dataset) == 10
    assert list(dataset.grp_list) == [f"grp{idx}" for idx in range(10)]

    sample = dataset[5]
    assert sorted(sample.keys()) == ["bar", "foo"]
    assert isinstance(sample["bar"], torch.Tensor)
    assert isinstance(sample["foo"], torch.Tensor)
    assert np.array_equal(sample["bar"].numpy(), f_hdf5["grp5"]["bar"][:])
    assert np.array_equal(sample["foo"].numpy(), f_hdf5["grp5"]["foo"][:])



import logging
import os

from lightning.fabric.plugins.environments.cluster_environment import ClusterEnvironment

log = logging.getLogger(__name__)


class ABCIEnvironment(ClusterEnvironment):
    """
    Environment class for ABCI.

    This class provides methods to interact with the ABCI environment,
    such as retrieving the world size, global rank, node rank, and local rank.
    """

    def __init__(self) -> None:
        self._world_size = int(os.environ["OMPI_COMM_WORLD_SIZE"])
        self._rank = int(os.environ["OMPI_COMM_WORLD_RANK"])
        self._local_rank = int(os.environ["OMPI_COMM_WORLD_LOCAL_RANK"])
        self._local_size = int(os.environ["OMPI_COMM_WORLD_LOCAL_SIZE"])

        self._main_address = os.environ["MAIN_ADDR"]
        self._main_port = int(os.environ["MAIN_PORT"])

    @property
    def creates_processes_externally(self) -> bool:
        return True

    @property
    def main_address(self) -> str:
        return self._main_address

    @property
    def main_port(self) -> int:
        return self._main_port

    @staticmethod
    def detect() -> bool:
        return True

    def world_size(self) -> int:
        return self._world_size

    def global_rank(self) -> int:
        return self._rank

    def node_rank(self) -> int:
        return self._rank // self._local_size

    def local_rank(self) -> int:
        return self._local_rank

    def set_world_size(self, size: int) -> None:
        if size != self.world_size():
            raise ValueError(f"`size` is expected to be {self.world_size()}, buf {size} is given.")

    def set_global_rank(self, rank: int) -> None:
        if rank != self.global_rank():
            raise ValueError(f"`rank` is expected to be {self.global_rank()}, buf {rank} is given.")

    def validate_settings(self, num_devices: int, num_nodes: int) -> None:
        if num_devices != self._local_size:
            raise ValueError("`num_devices` should match ${OMPI_COMM_WORLD_LOCAL_SIZE}")

        if num_devices * num_nodes != self._world_size:
            raise ValueError("`num_devices * num_nodes` should match ${OMPI_COMM_WORLD_SIZE}")



from collections.abc import Callable

import torch


class SequentialLR(torch.optim.lr_scheduler.SequentialLR):
    """
    A wrapper of torch.optim.lr_scheduler.SequentialLR to use list of functions
    to create schedulers.

    Args:
        optimizer: Optimizer.
        schedulers_fn: List of functions to create schedulers.
        milestones: List of epoch indices. Must be increasing.

    ... code-block:: yaml

        scheduler_generator:
          _partial_: True
          _convert_: "all"
          _target_: aiaccel.lr_schedulers.SequentialLR
          schedulers_fn:
            - _target_: torch.optim.lr_scheduler.LinearLR
              _partial_: True
              start_factor: 1.e-3
              end_factor: 1.0
              total_iters: 5000
            - _target_: torch.optim.lr_scheduler.CosineAnnealingLR
            _partial_: True
            T_max: 95000
          milestones: [5000]
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        schedulers_fn: list[Callable[[torch.optim.Optimizer], torch.optim.lr_scheduler._LRScheduler]],
        milestones: list[int],
    ):
        super().__init__(optimizer, [fn(optimizer) for fn in schedulers_fn], milestones)



from argparse import ArgumentParser

from hydra.utils import instantiate

from aiaccel.config import (
    load_config,
    overwrite_omegaconf_dumper,
    print_config,
    resolve_inherit,
)

overwrite_omegaconf_dumper()

parser = ArgumentParser()
parser.add_argument("config", type=str, help="Config file in YAML format")
args, unk_args = parser.parse_known_args()

config = load_config(args.config)
print_config(config)
config = resolve_inherit(config)

model = instantiate(config.model)

print(model)



from aiaccel.torch.datasets.cached_dataset import CachedDataset
from aiaccel.torch.datasets.file_cached_dataset import FileCachedDataset
from aiaccel.torch.datasets.hdf5_dataset import HDF5Dataset, RawHDF5Dataset
from aiaccel.torch.datasets.scatter_dataset import scatter_dataset

__all__ = [
    "CachedDataset",
    "FileCachedDataset",
    "RawHDF5Dataset",
    "HDF5Dataset",
    "scatter_dataset",
]



from typing import Any

from collections.abc import Callable

from optuna.trial import Trial

from aiaccel.hpo.optuna.hparams import Const, Float, Hparam, T


class HparamsManager:
    """
    Manages hyperparameters for optimization.
    This class allows defining hyperparameters with various types and provides
    a method to suggest hyperparameters for a given trial.
    Attributes:
        params (dict): A dictionary where keys are hyperparameter names and values
                       are callables that take a Trial object and return a hyperparameter value.
    Methods:
        __init__(**params_def: dict[str, int | float | str | list[int | float] | Hparam[T]]) -> None:
            Initializes the HparamsManager with the given hyperparameter definitions.
        suggest_hparams(trial: Trial) -> dict[str, float | int | str | list[float | int | str]]:
            Suggests hyperparameters for the given trial.
    """

    def __init__(self, **params_def: dict[str, int | float | str | list[int | float] | Hparam[T]]) -> None:
        self.params: dict[str, Callable[[Trial, str], Any]] = {}
        for name, param in params_def.items():
            if callable(param):
                self.params[name] = param
            else:
                if isinstance(param, list):
                    low, high = param
                    self.params[name] = Float(low=low, high=high)
                else:
                    self.params[name] = Const(value=param)

    def suggest_hparams(self, trial: Trial) -> dict[str, float | int | str | list[float | int | str]]:
        """
        Suggests hyperparameters for a given trial.
        This method generates a dictionary of hyperparameters by applying the
        parameter functions stored in `self.params` to the provided trial.
        Args:
            trial (Trial): An Optuna trial object used to suggest hyperparameters.
        Returns:
            dict[str, float | int | str | list[float | int | str]]: A dictionary
            where keys are parameter names and values are the suggested
            hyperparameters, which can be of type float, int, str, or a list of
            these types.
        """

        return {name: param_fn(trial, name) for name, param_fn in self.params.items()}



from typing import Any

from collections.abc import Callable
import copy
from importlib import resources
from pathlib import Path
import re

from colorama import Fore
from omegaconf import DictConfig, ListConfig, Node
from omegaconf import OmegaConf as oc  # noqa:N813
from omegaconf._utils import OmegaConfDumper

from simpleeval import simple_eval
import yaml
from yaml.resolver import BaseResolver


def overwrite_omegaconf_dumper(mode: str = "|") -> None:
    """
    Overwrites the default string representation in OmegaConf's YAML dumper.

    This function modifies the `OmegaConfDumper` to represent multi-line strings
    using the specified style (`mode`). By default, it uses the `|` block style
    for multi-line strings. Single-line strings remain unchanged.

    Args:
        mode (str, optional): The YAML style character for multi-line strings.
                              Defaults to "|".
    """

    def str_representer(dumper: OmegaConfDumper, data: str) -> yaml.Node:
        return dumper.represent_scalar(
            BaseResolver.DEFAULT_SCALAR_TAG, data, style=mode if len(data.splitlines()) > 1 else None
        )

    OmegaConfDumper.add_representer(str, str_representer)
    OmegaConfDumper.str_representer_added = True


def load_config(
    config_filename: str | Path,
    parent_config: dict[str, Any] | DictConfig | ListConfig | None = None,
) -> DictConfig | ListConfig:
    """Load YAML configuration

    When the user specifies ``_base_``, the specified YAML file is loaded as the base,
    and the original configuration is merged with the base config.
    If the configuration specified in ``_base_`` also contains ``_base_``, the process is handled recursively.

    Additionally, if `bootstrap_config` is provided, it is merged with the final
    configuration to ensure any default values or overrides are applied.

    Args:
        config (Path): Path to the configuration
        parent_config (dict[str, Any] | DictConfig | ListConfig | None):
            A configuration that is merged to the loaded configuration.
            This is intended to define default config paths (e.g., working_directory) dynamically.

    Returns:
        merge_user_config (DictConfig): The merged configuration of the base config and the original config
        user_config(DictConfig | ListConfig) : The configuration without ``_base_``

    """

    # Register custom resolvers
    oc.register_new_resolver("eval", simple_eval, replace=True)
    oc.register_new_resolver("resolve_pkg_path", resources.files, replace=True)

    if not isinstance(config_filename, Path):
        config_filename = Path(config_filename)

    if not config_filename.is_absolute():
        config_filename = Path.cwd() / config_filename

    if parent_config is None:
        parent_config = {}

    config = oc.merge(oc.load(config_filename), parent_config)

    if isinstance(config, DictConfig) and "_base_" in config:
        # process _base_
        base_paths = config["_base_"]
        if not isinstance(base_paths, ListConfig):
            base_paths = [base_paths]

        config.pop("_base_")
        for base_path in map(Path, base_paths):
            if not base_path.is_absolute():
                base_path = config_filename.parent / base_path

            config = load_config(base_path, config)

    return config


def print_config(config: ListConfig | DictConfig, line_length: int = 80) -> None:
    """
    Print the given configuration with syntax highlighting.

    This function converts `pathlib.Path` objects to strings before printing,
    ensuring that the output YAML format remains valid. It also highlights
    configuration keys in yellow for better readability.

    Args:
        config (ListConfig | DictConfig): The configuration to print.
        line_length (int, optional): The width of the separator line (default: 80).

    """

    config = pathlib2str_config(config)  # https://github.com/omry/omegaconf/issues/82

    print("=" * line_length)
    for line in oc.to_yaml(config).splitlines():
        print(re.sub(r"(\s*)(\w+):", rf"\1{Fore.YELLOW}\2{Fore.RESET}:", line, count=1))
    print("=" * line_length)


def apply_recursively(
    func: Callable[[DictConfig | ListConfig], DictConfig | ListConfig],
) -> Callable[[DictConfig | ListConfig], DictConfig | ListConfig]:
    """
    Recursively apply a function to all elements in a DictConfig or ListConfig.

    This function traverses the configuration structure and applies the provided
    function to each element, allowing for custom transformations or operations.

    Args:
        config (DictConfig | ListConfig): The configuration to traverse.
        func (callable): The function to apply to each element.

    Returns:
        DictConfig | ListConfig: The modified configuration with the function applied.
    """

    def _inner_fn(config: DictConfig | ListConfig) -> DictConfig | ListConfig:
        config = func(config)

        if isinstance(config, DictConfig):
            config = copy.deepcopy(config)

            for key in config:
                if (
                    isinstance(node_dict := config._get_node(key), Node)
                    and not node_dict._is_interpolation()
                    and isinstance(config[key], DictConfig | ListConfig)
                ):
                    config[key] = _inner_fn(config[key])
        elif isinstance(config, ListConfig):
            config = copy.deepcopy(config)

            for ii in range(len(config)):
                if (
                    isinstance(node_list := config._get_node(ii), Node)
                    and not node_list._is_interpolation()
                    and isinstance(config[ii], DictConfig | ListConfig)
                ):
                    config[ii] = _inner_fn(config[ii])

        return config

    return _inner_fn


@apply_recursively
def resolve_inherit(config: DictConfig | ListConfig) -> DictConfig | ListConfig:
    """Resolve _inherit_ in config

    Merge the dict in ``_inherit_`` into a dict of the same hierarchy.
    ``_inherit_`` is specified by omegaconf interpolation

    Args:
        config (DictConfig | ListConfig): The configuration loaded by load_config

    Returns:
        DictConfig | ListConfig: The configuration without ``_inherit_``
    """
    if isinstance(config, DictConfig) and "_inherit_" in config:
        inherit_configs = config["_inherit_"]
        if not isinstance(inherit_configs, ListConfig):
            inherit_configs = [inherit_configs]

        config.pop("_inherit_")

        for inherit_config in inherit_configs:
            if isinstance(inherit_config, DictConfig):
                config = oc.merge(inherit_config, config)

    return config


@apply_recursively
def pathlib2str_config(config: DictConfig | ListConfig) -> DictConfig | ListConfig:
    """
    Convert `pathlib.Path` objects in the configuration to strings.

    This function recursively traverses the configuration and replaces all `pathlib.Path`
    objects with their string representations. This is useful for saving the configuration
    in a YAML file, as YAML does not support `Path` objects.

    Args:
        config (ListConfig | DictConfig): The configuration to convert.

    Returns:
        ListConfig | DictConfig: The modified configuration with `Path` objects replaced by strings.

    """

    if isinstance(config, DictConfig):
        for k, v in config.items():
            if isinstance(v, Path):
                config[k] = str(v)

    return config



#! /usr/bin/env python


from pathlib import Path
import shlex
import subprocess
import time

from aiaccel.job.apps import prepare_argument_parser


def main() -> None:
    # Load configuration (from the default YAML string)
    config, parser, sub_parsers = prepare_argument_parser("sge.yaml")

    args = parser.parse_args()
    mode = args.mode + "-array" if getattr(args, "n_tasks", None) is not None else args.mode

    # Prepare the job script and arguments
    job = config[mode].job.format(command=shlex.join(args.command), args=args)

    if mode in ["cpu-array", "gpu-array"]:
        job = f"""\
for LOCAL_PROC_INDEX in {{1..{args.n_procs}}}; do
    TASK_INDEX=$(( SGE_TASK_ID + {args.n_tasks_per_proc} * (LOCAL_PROC_INDEX - 1) ))

    if [[ $TASK_INDEX -gt {args.n_tasks} ]]; then
        break
    fi

    TASK_INDEX=$TASK_INDEX \\
    TASK_STEPSIZE={args.n_tasks_per_proc} \\
        {job} > {args.log_filename.with_suffix("")}.${{SGE_TASK_ID}}-${{LOCAL_PROC_INDEX}}.log 2>&1 &

    pids[$LOCAL_PROC_INDEX]=$!
done

for i in "${{!pids[@]}}"; do
    wait ${{pids[$i]}}
done
"""
        job_log_filename = args.log_filename.with_suffix(".$TASK_ID.log")
        job_status_filename: Path = args.log_filename.with_suffix(".${SGE_TASK_ID}.out")

        status_filename_list = []
        for array_idx in range(0, args.n_tasks, args.n_tasks_per_proc * args.n_procs):
            status_filename_list.append(args.log_filename.with_suffix(f".{array_idx + 1}.out"))
    else:
        job_log_filename = args.log_filename
        job_status_filename = args.log_filename.with_suffix(".out")

        status_filename_list = [job_status_filename]

    job_script = f"""\
#! /bin/bash

#$-j y
#$-cwd
#$-o {job_log_filename}

set -eE -o pipefail
trap 'echo $? > {job_status_filename}' ERR EXIT  # at error and exit
trap 'echo 143 > {job_status_filename}' TERM  # at termination (by job scheduler)

if [ -n "$PBS_O_WORKDIR" ] && [ "$PBS_ENVIRONMENT" != "PBS_INTERACTIVE" ]; then
    cd $PBS_O_WORKDIR
fi


{config.script_prologue}

{job}
"""

    qsub = config.qsub.format(args=args)
    qsub_args = config[mode].qsub_args.format(args=args)

    # Create the job script file, remove old status files, and run the job
    args.log_filename.parent.mkdir(exist_ok=True, parents=True)

    job_filename: Path = args.log_filename.with_suffix(".sh")
    with open(job_filename, "w") as f:
        f.write(job_script)

    for status_filename in status_filename_list:
        status_filename.unlink(missing_ok=True)

    subprocess.run(f"{qsub} {qsub_args} {job_filename}", shell=True, check=True)

    for status_filename in status_filename_list:
        while not status_filename.exists():
            time.sleep(1.0)

        status = int(status_filename.read_text())
        if status != 0:
            raise RuntimeError(f"Job failed with {status} exit code.")
        status_filename.unlink()


if __name__ == "__main__":
    main()



import numpy.typing as npt
from typing import TypeVar

from collections.abc import Callable

import numpy as np

import torch.distributed as dist
from torch.utils.data import Dataset, Subset

T = TypeVar("T")


def scatter_dataset(
    dataset: Dataset[T],
    permute_fn: Callable[[npt.NDArray[np.int64]], npt.NDArray[np.int64]] | None = None,
) -> Subset[T]:
    """
    Splits a dataset into subsets and returns the subset corresponding to the current process rank.

    Args:
        dataset (Dataset[T]): The input dataset to be split.
        permute_fn (Callable[[npt.NDArray[np.int64]], npt.NDArray[np.int64]] | None, optional):
            A function that takes an array of indices and returns a permuted version of the array.
            If None, a default permutation function using np.random.Generator is used.
            Defaults to None.

    Returns:
        Subset[T]: The subset of the input dataset corresponding to the current process rank.
    """

    if permute_fn is None:
        permute_fn = np.random.Generator(np.random.PCG64(0)).permutation

    world_size = dist.get_world_size()
    rank = dist.get_rank()

    dataset_size = len(dataset)  # type: ignore[arg-type]
    total_size = int(np.ceil(dataset_size / world_size)) * world_size

    indices = permute_fn(np.arange(dataset_size))
    repeated_indices = np.concatenate([indices, indices[: total_size - dataset_size]])

    split_indices = np.split(repeated_indices, world_size)

    return Subset(dataset, list(split_indices[rank]))



import os
from unittest import mock

import pytest

from aiaccel.torch.lightning.abci_environment import ABCIEnvironment


@mock.patch.dict(
    os.environ,
    {
        "OMPI_COMM_WORLD_SIZE": "8",
        "OMPI_COMM_WORLD_RANK": "6",
        "OMPI_COMM_WORLD_LOCAL_RANK": "2",
        "OMPI_COMM_WORLD_LOCAL_SIZE": "4",
        "MAIN_ADDR": "192.168.0.1",
        "MAIN_PORT": "3000",
    },
)
def test_abci_environment() -> None:
    env = ABCIEnvironment()

    assert env.creates_processes_externally

    assert env.main_address == "192.168.0.1"
    assert env.main_port == 3000

    assert env.detect()
    assert env.world_size() == 8
    assert env.global_rank() == 6
    assert env.node_rank() == 1
    assert env.local_rank() == 2

    env.validate_settings(4, 2)

    with pytest.raises(ValueError, match=r"^`num_devices` should match.*"):
        env.validate_settings(3, 2)

    with pytest.raises(ValueError, match=r"^`num_devices \* num_nodes` should match.*"):
        env.validate_settings(4, 1)



#! /usr/bin/env python3


import logging
from math import ceil
from pathlib import Path
import shlex
import subprocess

from aiaccel.job.apps import prepare_argument_parser

logger = logging.getLogger(__name__)


def main() -> None:
    # Load configuration (from the default YAML string)
    config, parser, sub_parsers = prepare_argument_parser("local.yaml")

    args = parser.parse_args()
    mode = args.mode + "-array" if getattr(args, "n_tasks", None) is not None else args.mode

    for key in ["walltime", "n_nodes", "n_tasks_per_proc"]:
        if getattr(args, key, None) is not None:
            logger.warning(f"Argument '{key}' is defined for compatibility and will not be used in aiaccel-job local.")

    # Prepare the job script and arguments
    job = config[mode].job.format(command=shlex.join(args.command), args=args)

    if mode in ["cpu-array", "gpu-array"]:
        n_tasks_per_proc = ceil(args.n_tasks / args.n_procs)
        job = f"""\
for LOCAL_PROC_INDEX in {{1..{args.n_procs}}}; do
    TASK_INDEX=$(( 1 + {n_tasks_per_proc} * (LOCAL_PROC_INDEX - 1) ))

    if [[ $TASK_INDEX -gt {args.n_tasks} ]]; then
        break
    fi

    TASK_INDEX=$TASK_INDEX \\
    TASK_STEPSIZE={n_tasks_per_proc} \\
        {job} 2>&1 | tee {args.log_filename.with_suffix("")}.${{LOCAL_PROC_INDEX}}.log &

    pids[$LOCAL_PROC_INDEX]=$!
done

for i in "${{!pids[@]}}"; do
    wait ${{pids[$i]}}
done
"""
    else:
        job = f"{job} 2>&1 | tee {args.log_filename}"

    job_script = f"""\
#! /bin/bash

set -eE -o pipefail
trap 'exit $?' ERR EXIT  # at error and exit
trap 'echo 143' TERM  # at termination (by job scheduler)
trap 'kill 0' INT


{config.script_prologue}

{job}
"""

    # Create the job script file, remove old status files, and run the job
    args.log_filename.parent.mkdir(exist_ok=True, parents=True)

    job_filename: Path = args.log_filename.with_suffix(".sh")
    with open(job_filename, "w") as f:
        f.write(job_script)

    subprocess.run(f"bash {job_filename}", shell=True, check=True)


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
# Full options: https://www.sphinx-doc.org/en/master/usage/configuration.html

from pathlib import Path
import sys

import aiaccel

# -- Path setup --------------------------------------------------------------
root_path = Path(__file__).parent.parent.parent
sys.path.insert(0, str(root_path.absolute()))

# -- Project information -----------------------------------------------------
project = "aiaccel"
author = "National Institute of Advanced Industrial Science And Technology (AIST)"
project_copyright = author
release = aiaccel.__version__

html_logo = f"{root_path}/docs/image/logo_aiaccel.png"
html_favicon = f"{root_path}/docs/image/favicon.ico"

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx_design",
    "sphinx_copybutton",
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

language = "en"

# -- HTML output -------------------------------------------------------------
html_theme = "shibuya"
html_show_sourcelink = False
html_show_sphinx = False

html_static_path = ["_static"]
html_theme_options = {
    "accent_color": "gold",
    "nav_links": [
        {"title": "User Guide", "url": "user_guide/index"},
        {
            "title": "API Reference",
            "url": "api_reference/index",
            "children": [
                {
                    "title": "OmegaConf Utilities",
                    "url": "api_reference/config",
                    "summary": "aiaccel.config",
                },
                {
                    "title": "PyTorch/Lightning Toolkit",
                    "url": "api_reference/torch",
                    "summary": "aiaccel.torch",
                },
                {
                    "title": "Hyperparameter Optimization",
                    "url": "api_reference/hpo",
                    "summary": "aiaccel.hpo",
                },
            ],
        },
        {"title": "Contribution Guide", "url": "contribution_guide/index"},
    ],
    "github_url": "https://github.com/aistairc/aiaccel",
    "globaltoc_expand_depth": 1,
}



from typing import Any

from collections.abc import Callable

from torch.utils.data import DataLoader, Dataset

import lightning as lt

from aiaccel.torch.datasets import CachedDataset, scatter_dataset


class SingleDataModule(lt.LightningDataModule):
    """
    A PyTorch Lightning DataModule designed to handle training and validation datasets
    with support for caching and dataset scattering.

    Attributes:
        train_dataset_fn (Callable[..., Dataset[str]]): A callable function to create the training dataset.
        val_dataset_fn (Callable[..., Dataset[str]]): A callable function to create the validation dataset.
        batch_size (int): The batch size for the DataLoader.
        use_cache (bool): Whether to cache the datasets. Defaults to False.
        use_scatter (bool): Whether to scatter the datasets. Defaults to True.
        num_workers (int): Number of workers for the DataLoader. Defaults to 10.
        common_args (dict[str, Any] | None): Common arguments to pass to the dataset functions. Defaults to None.
    Methods:
        setup(stage: str | None) -> None:
            Prepares the datasets for training and validation. Only supports the "fit" stage.
            Raises a ValueError if the stage is not "fit".
        train_dataloader() -> DataLoader:
            Returns the DataLoader for the training dataset.
        val_dataloader() -> DataLoader:
            Returns the DataLoader for the validation dataset.
        _create_dataloader(dataset, **kwargs: Any) -> DataLoader:
            Internal method to create a DataLoader for a given dataset with specified configurations.
    """

    def __init__(
        self,
        train_dataset_fn: Callable[..., Dataset[str]],
        val_dataset_fn: Callable[..., Dataset[str]],
        batch_size: int,
        use_cache: bool = False,
        use_scatter: bool = True,
        num_workers: int = 10,
        common_args: dict[str, Any] | None = None,
    ):
        super().__init__()

        self.train_dataset_fn = train_dataset_fn
        self.val_dataset_fn = val_dataset_fn

        self.common_args = common_args if common_args is not None else {}

        self.batch_size = batch_size

        self.use_cache = use_cache
        self.use_scatter = use_scatter

        self.num_workers = num_workers

    def setup(self, stage: str | None) -> None:
        if stage == "fit":
            train_dataset = self.train_dataset_fn(**self.common_args)
            val_dataset = self.val_dataset_fn(**self.common_args)

            print(f"Dataset size: {len(train_dataset)=},  {len(val_dataset)=}")  # type: ignore

            if self.use_cache:
                train_dataset = CachedDataset(train_dataset)
                val_dataset = CachedDataset(val_dataset)

            if self.use_scatter:
                train_dataset = scatter_dataset(train_dataset)
                val_dataset = scatter_dataset(val_dataset)

            self.train_dataset = train_dataset
            self.val_dataset = val_dataset
        else:
            raise ValueError("`stage` is not 'fit'.")

    def _create_dataloader(self, dataset: Dataset[Any], **kwargs: Any) -> DataLoader[Any]:
        return DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=True,
            shuffle=True,
            pin_memory=True,
            **kwargs,
        )

    def train_dataloader(self) -> DataLoader[Any]:
        return self._create_dataloader(self.train_dataset, drop_last=True)

    def val_dataloader(self) -> DataLoader[Any]:
        return self._create_dataloader(self.val_dataset, drop_last=False)



import torch
from torch import nn
from torch.nn import functional as fn

from torchmetrics.classification import MulticlassAccuracy

from aiaccel.torch.lightning import OptimizerConfig, OptimizerLightningModule


class ImageClassificationTask(OptimizerLightningModule):
    def __init__(self, model: nn.Module, optimizer_config: OptimizerConfig, num_classes: int = 10):
        super().__init__(optimizer_config)

        self.model = model

        self.training_accuracy = MulticlassAccuracy(num_classes=num_classes)
        self.validation_accuracy = MulticlassAccuracy(num_classes=num_classes)

    @torch.compile
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)  # type: ignore

    def training_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> torch.Tensor:
        x, y = batch

        logits = self(x)

        loss = fn.cross_entropy(logits, y)

        self.log_dict(
            {
                "training/loss": loss,
                "training/accuracy": self.training_accuracy(logits, y),
            },
            prog_bar=True,
        )

        return loss

    def validation_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> None:
        x, y = batch

        logits = self(x)

        loss = fn.cross_entropy(logits, y)

        self.log_dict(
            {
                "validation/loss": loss,
                "validation/accuracy": self.validation_accuracy(logits, y),
            },
            prog_bar=True,
        )



import glob

from matplotlib.axes._axes import Axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def get_min_values(values: list[float]) -> list[float]:
    min_values = []
    min_value = np.inf
    for value in values:
        if value < min_value:
            min_value = value
        min_values.append(min_value)

    return min_values


def plot_dim_vs_min_value(ax: Axes, csv_names_for_dim: list[list[list[str]]], title: str) -> None:
    label_names = ["nm", "TPE-mcT-mvF", "nm+subTPE-mcT-mvF"]
    colors = ["r", "g", "b"]

    dim = ["2", "3", "5", "10", "20", "40"]

    # dim
    min_values_for_dim = []
    for csv_names_for_algorism in csv_names_for_dim:
        # nm, tpe, nm_subtpe
        min_values_for_algorism = []
        for csv_names in csv_names_for_algorism:
            min_values = []
            for csv_name in csv_names:
                df = pd.read_csv(csv_name)

                df_values = df["value - f_opt"]
                print(df_values)
                print(min(df_values))
                min_values.append(min(df_values))
            min_values_for_algorism.append(min_values)
        min_values_for_dim.append(min_values_for_algorism)

    print(len(min_values_for_dim))

    for i in range(3):
        values = [item[i] for item in min_values_for_dim]
        print(values)
        values_mean = np.array(values).mean(axis=1)
        values_std = np.array(values).std(axis=1)
        ax.errorbar(
            dim,
            values_mean,
            yerr=values_std,
            capsize=5,
            markersize=10,
            ecolor=colors[i],
            markeredgecolor=colors[i],
            color=colors[i],
            label=label_names[i],
        )

    if title == "f6":
        ax.set_ylim(-1000, 10000)
    ax.set_title(title)
    ax.grid(axis="both")
    ax.legend(fontsize=6)


def compare_optimizer(base_dir: str = ".") -> None:
    fig, ax = plt.subplots(5, 5, figsize=(16, 20))
    if isinstance(ax, Axes):
        return

    for num_of_f in range(1, 25):
        result_csv_list_for_dim = []
        for num_of_dm in [2, 3, 5, 10, 20, 40]:
            result_csv_patterns = [
                f"{base_dir}/nelder-mead/optuna_csv/optuna-nelder-mead-func_id{num_of_f}-dim{num_of_dm}-instance*/f{num_of_f}/DM{num_of_dm:02}/result_bbob_f{num_of_f:03}_i*_d{num_of_dm:02}_*_fopt.csv",
                f"{base_dir}/TPE/optuna_csv/optuna-TPE-func_id{num_of_f}-dim{num_of_dm}-instance*/f{num_of_f}/DM{num_of_dm:02}/result_bbob_f{num_of_f:03}_i*_d{num_of_dm:02}_*_fopt.csv",
                f"{base_dir}/nelder-mead-subTPE/optuna_csv/optuna-nelder-mead-subTPE-func_id{num_of_f}-dim{num_of_dm}-instance*/f{num_of_f}/DM{num_of_dm:02}/result_bbob_f{num_of_f:03}_i*_d{num_of_dm:02}_*_fopt.csv",
            ]

            result_csv_list = [sorted(glob.glob(pattern)) for pattern in result_csv_patterns]
            print(result_csv_list)
            result_csv_list_for_dim.append(result_csv_list)

        plot_dim_vs_min_value(
            ax[int((num_of_f - 1) / 5), int((num_of_f - 1) % 5)], result_csv_list_for_dim, f"f{num_of_f}"
        )

    plt.savefig("result_bbob_dim_vs_value-fopt_parallel.png")
    plt.show()


if __name__ == "__main__":
    compare_optimizer()



from concurrent.futures import ThreadPoolExecutor
from itertools import product
import subprocess


def main() -> None:
    sampler_names = ["nelder-mead", "nelder-mead-subTPE", "TPE"]
    func_ids = list(range(1, 25))
    dims = [2, 3, 5, 10, 20, 40]
    execute_times = ["0:01:00", "0:02:00", "0:03:00", "0:10:00", "0:30:00", "3:00:00"]
    instances = list(range(1, 16))
    optuna_seeds = list(range(1, 16))

    combinations = product(
        sampler_names, func_ids, zip(dims, execute_times, strict=False), zip(instances, optuna_seeds, strict=False)
    )

    with ThreadPoolExecutor() as pool:
        for sampler_name, func_id, (dim, execute_time), (instance, optuna_seed) in combinations:
            execute_time = "0:05:00" if sampler_name == "nelder-mead" else execute_time
            print(sampler_name, (func_id, execute_time), dim, (instance, optuna_seed))

            aiaccel_job_command = f"""\
aiaccel-job pbs --config job_config.yaml cpu --walltime {execute_time} log/job_{func_id}_{dim}_{instance}.log \
-- python3.13 experiment_coco.py --func_id {func_id} --dim {dim} \
--instance {instance} --optuna_seed {optuna_seed} --sampler_name {sampler_name}
"""

            pool.submit(subprocess.run, aiaccel_job_command, shell=True)


if __name__ == "__main__":
    main()



from collections import defaultdict

from torch.utils.data import Dataset

from aiaccel.torch.datasets.cached_dataset import CachedDataset


def test_cached_dataset() -> None:
    class DummyDataset(Dataset[int]):
        def __init__(self) -> None:
            self.counter = defaultdict[int, int](lambda: 0)

        def __getitem__(self, index: int) -> int:
            self.counter[index] += 1
            return index

    orig_dataset = DummyDataset()
    dataset = CachedDataset(orig_dataset)

    for _ in range(2):
        for ii in range(5):
            assert dataset[ii] == ii

    assert all(count == 1 for count in orig_dataset.counter.values())



from collections.abc import Callable, Generator
from contextlib import AbstractContextManager, contextmanager
import os
from pathlib import Path
import shutil
import subprocess

from hydra.utils import instantiate

import pytest

from aiaccel.config import load_config


@pytest.fixture()
def workspace_factory(
    tmp_path_factory: pytest.TempPathFactory,
) -> Callable[[str], AbstractContextManager[Path]]:
    @contextmanager
    def _factory(data_name: str = "single_objective") -> Generator[Path, None, None]:
        tmp_path = tmp_path_factory.mktemp("workspace")

        shutil.copytree(Path(__file__).parent / "data" / data_name, tmp_path, dirs_exist_ok=True)

        org_path = Path.cwd()

        try:
            os.chdir(tmp_path)
            yield tmp_path
        finally:
            os.chdir(org_path)

    return _factory


def test_from_config(workspace_factory: Callable[..., AbstractContextManager[Path]]) -> None:
    with workspace_factory() as workspace:
        subprocess.run("aiaccel-hpo optimize --config=config.yaml", shell=True, check=True)

        assert (workspace / "optuna.db").exists()

        config = load_config(workspace / "merged_config.yaml")
        study = instantiate(config.study)
        assert len(study.get_trials()) == 15


def test_from_cli(workspace_factory: Callable[..., AbstractContextManager[Path]]) -> None:
    with workspace_factory() as workspace:
        cmd = (
            "aiaccel-hpo optimize"
            " working_directory=./cli/"
            " n_trials=15"
            " n_max_jobs=1"
            " params.x1='[0,1]'"
            " params.x2='[0,1]'"
            " study.sampler._target_=optuna.samplers.TPESampler"
            " study.sampler.seed=0"
            " --"
            " python ./objective.py --x1={x1} --x2={x2} {out_filename}"
        )
        subprocess.run(cmd, shell=True, check=True)

        config = load_config(workspace / "cli" / "merged_config.yaml")
        study = instantiate(config.study)
        best_value = study.best_trial.value

    # check consistency with the config-style execution
    with workspace_factory() as workspace:
        subprocess.run("aiaccel-hpo optimize --config=config.yaml", shell=True, check=True)

        config = load_config(workspace / "merged_config.yaml")
        study = instantiate(config.study)

    assert best_value == study.best_trial.value


def test_resume(workspace_factory: Callable[..., AbstractContextManager[Path]]) -> None:
    with workspace_factory() as workspace:
        subprocess.run("aiaccel-hpo optimize --config=config.yaml", shell=True, check=True)
        subprocess.run("aiaccel-hpo optimize --config=config.yaml", shell=True, check=True)

        config = load_config(workspace / "merged_config.yaml")
        study = instantiate(config.study)
        assert len(study.get_trials()) == 30


def test_multi_objective(workspace_factory: Callable[..., AbstractContextManager[Path]]) -> None:
    with workspace_factory("multi_objective") as workspace:
        subprocess.run("aiaccel-hpo optimize --config=config.yaml", shell=True, check=True)

        config = load_config(workspace / "merged_config.yaml")
        study = instantiate(config.study)

        assert len(study.get_trials()) == 15

        assert all(len(trial.values) == 2 for trial in study.get_trials())


def test_from_cli_and_config(workspace_factory: Callable[..., AbstractContextManager[Path]]) -> None:
    with workspace_factory() as workspace:
        cmd = (
            "aiaccel-hpo optimize"
            " n_trials=30"
            " n_max_jobs=1"
            " params.x1='[0,10]'"
            " params.x2='[0,10]'"
            " study.sampler._target_=optuna.samplers.TPESampler"
            " study.sampler.seed=0"
            " --config=config.yaml"
            " --"
        )
        subprocess.run(cmd, shell=True, check=True)

        assert (workspace / "optuna.db").exists()

        config = load_config(workspace / "merged_config.yaml")
        study = instantiate(config.study)
        assert len(study.get_trials()) == 30
